// -*- mode: c++; c-indent-level: 4; c++-member-init-indent: 8; c++-electric-colon: 1; comment-column: 35; -*-
// GENVAR.CC : implementation of gene-class functions

#include <sys/types.h>

// header file for this one...
#include "genvar.hpp"

const unsigned LENGTH_0 = 1, SIZE_0 = 2;

// -------------------------------------------------------------
void genVar::initSize( unsigned _sizeAllele ) {
// -- Checks and asigns size and related vars
// -------------------------------------------------------------

  if (!_sizeAllele) {
    cerr << "Alleles of zero length\n";
    exit( SIZE_0 );
  }

  size = _sizeAllele;
  maxForSize = (unsigned long) pow(MAX8BITS, size);
}


// ---------------------------------------------------------------------
genVar::genVar(  genLength_t _numAlleles, unsigned _sizeAllele )
    :rawGene( _numAlleles*_sizeAllele ) {
// -- usual initial population constructor, or for tests
// ---------------------------------------------------------------------

    initSize( _sizeAllele );

}


// ---------------------- Constructor with 2 parents -------------------


// --------------------------------------------------------------------
genVar::genVar(const genVar & _genVar1, 
	       const genVar & _genVar2,
	       xOverRate_t _xOverRate, mutRate_t _mutationRate, 
	       killRate_t _killRate, incRate_t _dupRate)
:rawGene( _genVar1, _genVar2, _xOverRate, _mutationRate ) 
//  --	Builds a genome by mutating genVar1, with given MutationRate,
//	and adding genomes according to the proportions
// ---------------------------------------------------------------------

{
  initSize( _genVar1.getSize() );
	
  unsigned auxL = _genVar1.getLength(); 	    // computes length
  unsigned auxL2 = _genVar2.getLength();
  
  if ( auxL2 < auxL )			    // take smaller one
    auxL = auxL2;
  
  if ( myrand(10000)/10000.0 < _dupRate)
    dupAllele( (genLength_t) myrand( auxL ), _mutationRate );
  
  if ( myrand(10000)/10000.0 < _killRate)
    killAllele(  (genLength_t) myrand( auxL ) );
}


// --------------------------------------------------------------------
genVar::genVar(const genVar & _genVar1, 
	       mutRate_t _mutationRate, 
	       killRate_t _killRate, 
	       incRate_t _dupRate)
:rawGene( _genVar1, _mutationRate ){
//	--  Builds a genVarome by mutating genVare1, with given MutationRate
// ---------------------------------------------------------------------

  unsigned auxL = _genVar1.getLength(); 		// computes length
	
  initSize( _genVar1.getSize() );
  
  if ( myrand(10000)/10000.0 < _dupRate)
    dupAllele( (genLength_t) myrand( auxL ), _mutationRate );
  
  if ( myrand(10000)/10000.0 < _killRate)
    killAllele( (genLength_t) myrand( auxL ) );
}



// --------------------------------------------------------------------
genVar::genVar(const genVar& _genVar1,  
	       genOpMode _modeOp, 
	       genLength_t _geneIdx, 
	       mutRate_t _mutationRate,
	       genLength_t _geneIdx2 ):
	rawGene( _genVar1, _mutationRate )	{
// --	creates a new genome using unary genetic operators; uses vars as
//	needed, _mutationRate is used always. _geneIdx2 is used only
//      for transposition
// ---------------------------------------------------------------------



    // all instance variables get initialized
    initSize( _genVar1.getSize() );

    // now applies unary genetic operator as specified to the newly
    // created genome. If mode == CLONE, do nothing; already done.
    switch( _modeOp ) {
	case DUP :
	    dupAllele( _geneIdx, _mutationRate );
	    break;
	case KILL :
	    killAllele( _geneIdx );
	    break;
	case RANDINC :
	    addAllele();
	    break;
	case TRANSP :
	    if ( _geneIdx != _geneIdx2 ) 
		transposeChunk(_geneIdx * BITSINBYTE * size, 
			       _geneIdx2 * BITSINBYTE * size,
			       size*BITSINBYTE);
	    break;	
       default:			   // To handle clone, and avoid
	    break;		   // warnings

    }
   
}



// ---------------------------------------------------------------------
void  genVar::getAllele( genLength_t idx, char* _inStr ) const {
// ---------------------------------------------------------------------
	memcpy( _inStr, getGenome() + idx*size, size );
}


// ---------------------------------------------------------------------
const char * genVar::getAllele( genLength_t idx ) const {
// getAllele- gets a string of size --size--
// ---------------------------------------------------------------------
    return  getGenome() + idx*size;
}


// ---------------------------------------------------------------------
void genVar::addAllele( ) {
//  --  adds another allele (size bytes) at the end of the string, of si
//      ze fixed by the instance variable
// ---------------------------------------------------------------------
    addChunkEnd( size );
}


// ---------------------------------------------------------------------
void genVar::killAllele(  genLength_t _pos ){
// -- deletes an allele from position idx, redims string
// ---------------------------------------------------------------------
    killChunk( size, _pos*size );			 
}


// ---------------------------------------------------------------------
void genVar::dupAllele(genLength_t _pos, 
		       mutRate_t _mutationRate ) {
//  -- duplicates the given allele with errors
// ---------------------------------------------------------------------

    dupChunk( size, _pos*size );
    for ( int i = 0; i < size*BITSINBYTE; i ++ ) 
      // now alters a little bit the new gene
      if ( ( myrand(100)/100.0 ) < _mutationRate)
	changeBit( _pos*size*BITSINBYTE + i );

}


// -----------------------------------------------------------------------
// Friendly operators
// ------------------

// -----------------------------------------------------------------------
ostream& operator << ( ostream& s, const genVar & _inGen ) {
// --	to print a genVar to stdout
// -----------------------------------------------------------------------

    for ( genLength_t i = 0; i < _inGen.getLength(); i ++ ) 
	s << (int) (unsigned char) _inGen.getStr(i) << " ";
	
    return s;
}


// ------------------------------------------------------------------------
// Non-Class funcs.
// _______________________________________________________________________

// --------------------------------------------
unsigned char deGray( unsigned char _deG ) {
// --	Pases from Gray Code to the decimal
//	representation. The formula is
//	y7 = x7
//	y6 = x7 ^ x6... and so on
// ---------------------------------------------
    
    unsigned char outG = _deG & Wgts[1];
    unsigned char thisBit = outG;
    for ( unsigned char i = 2; i <= 8; i ++ ) {
#ifdef DEBUG
	unsigned char prevOutG = outG;
#endif
	thisBit = (thisBit >> 1 ) ^ (_deG & Wgts[i]);
	outG |= thisBit;
#ifdef DEBUG
	printf( "%d %d %d %o %o %o\n",  i,(prevOutG >> 1 ),  (_deG & Wgts[i]), outG, _deG, Wgts[i] );
#endif
    }

    return outG;
}
